package org.processmining.plugins.timefilter.ui;

import java.awt.Cursor;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.lang.ref.SoftReference;
import java.util.ArrayList;
import java.util.Date;
import java.util.concurrent.TimeUnit;

import javax.swing.JPanel;

import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.timefilter.LogFilter;

/**
 * This class contains the widget with the lined chart
 * 
 * @author Andrea Burattin
 */
public class LinedChart 
	extends JPanel
	implements MouseMotionListener, MouseListener {

	/**
	 * Possible sorting options of the representation of the log 
	 */
	public enum SORTING {
		BY_DURATION,
		BY_STARTING_EVENT
	};
	
	// graphic stuff
	private static final long serialVersionUID = 8008300838558681660L;
	protected SoftReference<BufferedImage> bufferLineChartLayer = null;
	protected SoftReference<BufferedImage> bufferHideBarsLayer = null;
	private int height;
	private int width;
	private int internalMargin = 10;
	private int verticalLinesNumber = 5;
	private int traceBarHeight = 2;
	private int dividerBarSize = 2;
	private int dividerPointSize = 10;
	private boolean draggingTopBar = false;
	private boolean draggingBottomBar = false;
	private Font defaultFont = null;
	private FontMetrics defaultFontMetric = null;
	
	// logic stuff
	private LogFilter logFilter;
	private XLogInfo logInfo;
	private SORTING sorting;
	private ArrayList<XTrace> elements;
	private int topDivider = 0;
	private int bottomDivider = 0;
	
	// internal stuff
	long timeFirstEvent;
	long timeLastEvent;

	/**
	 * Chart constructor
	 * 
	 * @param logFilter a reference to the main UI component
	 * @param elements array of traces
	 * @param logInfo
	 * @param sorting
	 */
	public LinedChart(LogFilter logFilter, ArrayList<XTrace> elements, XLogInfo logInfo, SORTING sorting) {
		setOpaque(false);
		
		this.logFilter = logFilter;
		this.elements = elements;
		this.logInfo = logInfo;
		this.sorting = sorting;
		
		timeFirstEvent = logInfo.getLogTimeBoundaries().getStartDate().getTime();
		timeLastEvent = logInfo.getLogTimeBoundaries().getEndDate().getTime();
		
		addMouseMotionListener(this);
		addMouseListener(this);
	}
	
	/**
	 * 
	 * @param sorting
	 */
	public void changeSorting(ArrayList<XTrace> elements, SORTING sorting) {
		this.elements = elements;
		this.sorting = sorting;
		this.bufferLineChartLayer = null;
		
		// reset operations
		topDivider = 0;
		bottomDivider = 0;
		logFilter.updateFrequencies(getLowerBound(), getUpperBound());
	}
	
	/**
	 * 
	 * @return
	 */
	public int getLowerBound() {
		int elementsCount = elements.size();
		int fixedHeight = height - internalMargin*2;
		double position = (double) topDivider / (double) fixedHeight;
		return (int) Math.round(position * elementsCount);
	}
	
	/**
	 * 
	 * @return
	 */
	public int getUpperBound() {
		int elementsCount = elements.size();
		int fixedHeight = height - internalMargin*2;
		double position = (double) bottomDivider / (double) fixedHeight;
		return (int) Math.round(elementsCount - (position * elementsCount));
	}
	
	@Override
	protected void paintComponent(Graphics g) {
		height = this.getHeight();
		width = this.getWidth();
		Rectangle clip = g.getClipBounds();
		
		// first layer (line chart) --------------------------------------------
		
		if (bufferLineChartLayer == null) {
			
			// create new back buffer
			bufferLineChartLayer = new SoftReference<BufferedImage>(new BufferedImage(width, height, BufferedImage.TRANSLUCENT));
			Graphics2D g2d = bufferLineChartLayer.get().createGraphics();
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			
			// get the font
			if (defaultFont == null) {
				defaultFont = g2d.getFont().deriveFont(10f);
				defaultFontMetric = g2d.getFontMetrics(defaultFont);
			}
			
			// background
			g2d.setColor(UIColors.lineChartBackground);
			g2d.fillRoundRect(0, 0, width - 1, height - 1, 15, 15);
			
			// vertical lines
			g2d.setColor(UIColors.lineChartVerticalBars);
			int spaceBetweenLines = width / (verticalLinesNumber + 1);
			for (int i = 1; i <= verticalLinesNumber; i++) {
				g2d.drawLine(spaceBetweenLines*i, internalMargin, spaceBetweenLines*i, height-internalMargin);
			}
			
			// calculate the number of spot per pixel
			long longestTrace = -1;
			double millisecRatio = (double)(width - (internalMargin * 2)) / (timeLastEvent - timeFirstEvent);
			double verticalSpace = (double) (height - (internalMargin * 2)) / logInfo.getNumberOfTraces();
			
			if (sorting.equals(SORTING.BY_DURATION)) {
				longestTrace = -1;
				for (XTrace t : elements) {
					long traceFirst = logInfo.getTraceTimeBoundaries(t).getStartDate().getTime();
					long traceLast = logInfo.getTraceTimeBoundaries(t).getEndDate().getTime();
					if (longestTrace < (traceLast - traceFirst)) {
						longestTrace = traceLast - traceFirst;
					}
				}
				millisecRatio = (double)(width - (internalMargin * 2)) / longestTrace;
			}
			
			int row = 1;
			int x = 0;
			int y = 0;
			
			// draw the actual traces
			g2d.setColor(UIColors.lineChartBarColorUniform);
			for (XTrace t : elements) {
				long traceFirst = logInfo.getTraceTimeBoundaries(t).getStartDate().getTime();
				long traceLast = logInfo.getTraceTimeBoundaries(t).getEndDate().getTime();
				int w = (int) (millisecRatio * (traceLast - traceFirst));
				
				if (sorting.equals(SORTING.BY_DURATION)) {
					// trace color
					double value = (double)w / (double)(width - internalMargin * 2);
					g2d.setColor(UIColors.fromValueToColor(value));
				}
				
				// coordinates
				y = (int) (row++ * verticalSpace) + internalMargin - (traceBarHeight / 2) - (int) (verticalSpace / 2);
				if (sorting.equals(SORTING.BY_DURATION)) {
					x = internalMargin;
				} else if (sorting.equals(SORTING.BY_STARTING_EVENT)) {
					x = (int) (millisecRatio * (traceFirst - timeFirstEvent)) + internalMargin;
				}
				
				g2d.fillRoundRect(x, y, w, traceBarHeight, 20, 20);
			}
			
			// add texts
			g2d.setColor(UIColors.lineChartTextcolor);
			g2d.setFont(defaultFont);
			if (sorting.equals(SORTING.BY_DURATION)) {
				int days = (int) TimeUnit.DAYS.convert(longestTrace, TimeUnit.MILLISECONDS);
				int hours = (int) TimeUnit.HOURS.convert(longestTrace - (days*86400000L), TimeUnit.MILLISECONDS);
				int minutes = (int) TimeUnit.MINUTES.convert(longestTrace - (days*86400000L) - (hours*3600000L), TimeUnit.MILLISECONDS);
				int seconds = (int) TimeUnit.SECONDS.convert(longestTrace - (days*86400000L) - (hours*3600000L) - (minutes*60000), TimeUnit.MILLISECONDS);
				String duration = days + "d " + hours + "h " + minutes + "m " + seconds + "s";
				g2d.drawString(duration, width - defaultFontMetric.stringWidth(duration) - 10, 30);
			} else if (sorting.equals(SORTING.BY_STARTING_EVENT)) {
				String lastEvent = new Date(timeLastEvent).toString();
				g2d.drawString(new Date(timeFirstEvent).toString(), 10, 20);
				g2d.drawString(lastEvent, width - defaultFontMetric.stringWidth(lastEvent) - 10, 20);
				g2d.drawString(lastEvent, width - defaultFontMetric.stringWidth(lastEvent) - 10, height - 15);
			}
			
			// add legend
			g2d.setColor(UIColors.lineChartLegend);
			if (sorting.equals(SORTING.BY_DURATION)) {
				
				String labelAxisX = "Duration";
				String labelAxisY = "Traces";
				int spaceFromBorderX = 10;
				int spaceFromBorderY = 20;
				int verticalLineLenght = 15 + defaultFontMetric.stringWidth(labelAxisY);
				int horizontalLineLenght = 10 + defaultFontMetric.stringWidth(labelAxisX);
				g2d.drawLine(
						width - internalMargin - spaceFromBorderX - horizontalLineLenght, height - internalMargin - spaceFromBorderY - 1,
						width - internalMargin - spaceFromBorderX - horizontalLineLenght, height - internalMargin - spaceFromBorderY - verticalLineLenght);
				g2d.fillPolygon(new Polygon(new int[]{
						width - internalMargin - spaceFromBorderX - horizontalLineLenght - 2,
						width - internalMargin - spaceFromBorderX - horizontalLineLenght + 3,
						width - internalMargin - spaceFromBorderX - horizontalLineLenght
					}, new int[] {
						height - internalMargin - spaceFromBorderY - verticalLineLenght,
						height - internalMargin - spaceFromBorderY - verticalLineLenght,
						height - internalMargin - spaceFromBorderY - verticalLineLenght - 7
					}, 3));
				g2d.drawLine(
						width - internalMargin - spaceFromBorderX - horizontalLineLenght, height - internalMargin - spaceFromBorderY,
						width - internalMargin - spaceFromBorderX, height - internalMargin - spaceFromBorderY);
				g2d.fillPolygon(new Polygon(new int[]{
						width - internalMargin - spaceFromBorderX + 1,
						width - internalMargin - spaceFromBorderX + 7,
						width - internalMargin - spaceFromBorderX + 1
					}, new int[] {
						height - internalMargin - spaceFromBorderY - 2,
						height - internalMargin - spaceFromBorderY,
						height - internalMargin - spaceFromBorderY + 3
					}, 3));
				g2d.rotate(Math.toRadians(-90));
				g2d.drawString(
						labelAxisY,
						-(height - internalMargin - spaceFromBorderY - 5),
						width - internalMargin - spaceFromBorderX - horizontalLineLenght - defaultFontMetric.getDescent());
				g2d.rotate(Math.toRadians(90));
				g2d.drawString(
						labelAxisX,
						width - internalMargin - spaceFromBorderX - horizontalLineLenght + 5,
						height - internalMargin - spaceFromBorderY + defaultFontMetric.getAscent());
				
			} else if (sorting.equals(SORTING.BY_STARTING_EVENT)) {
				
				String labelAxisX = "Starting time";
				String labelAxisY = "Traces";
				int spaceFromBorderX = 10;
				int spaceFromBorderY = 20;
				int verticalLineLenght = 15 + defaultFontMetric.stringWidth(labelAxisY);
				int horizontalLineLenght = 10 + defaultFontMetric.stringWidth(labelAxisX);
				g2d.drawLine(
						internalMargin + spaceFromBorderX, height - internalMargin - spaceFromBorderY - 1,
						internalMargin + spaceFromBorderX, height - internalMargin - spaceFromBorderY - verticalLineLenght);
				g2d.fillPolygon(new Polygon(new int[]{
						internalMargin + spaceFromBorderX - 2,
						internalMargin + spaceFromBorderX + 3,
						internalMargin + spaceFromBorderX
					}, new int[] {
						height - internalMargin - spaceFromBorderY - verticalLineLenght,
						height - internalMargin - spaceFromBorderY - verticalLineLenght,
						height - internalMargin - spaceFromBorderY - verticalLineLenght - 7
					}, 3));
				g2d.drawLine(
						internalMargin + spaceFromBorderX, height - internalMargin - spaceFromBorderY,
						internalMargin + spaceFromBorderX + horizontalLineLenght, height - internalMargin - spaceFromBorderY);
				g2d.fillPolygon(new Polygon(new int[]{
						internalMargin + spaceFromBorderX + horizontalLineLenght + 1,
						internalMargin + spaceFromBorderX + horizontalLineLenght + 7,
						internalMargin + spaceFromBorderX + horizontalLineLenght + 1
					}, new int[] {
						height - internalMargin - spaceFromBorderY - 2,
						height - internalMargin - spaceFromBorderY,
						height - internalMargin - spaceFromBorderY + 3
					}, 3));
				g2d.rotate(Math.toRadians(-90));
				g2d.drawString(
						labelAxisY,
						-(height - internalMargin - spaceFromBorderY - 5),
						internalMargin + spaceFromBorderX - defaultFontMetric.getDescent());
				g2d.rotate(Math.toRadians(90));
				g2d.drawString(
						labelAxisX,
						internalMargin + spaceFromBorderX + 5,
						height - internalMargin - spaceFromBorderY + defaultFontMetric.getAscent());
			}
			
			g2d.dispose();
		}
		
		// second layer (hiding bars) ------------------------------------------
		
		bufferHideBarsLayer = new SoftReference<BufferedImage>(new BufferedImage(width, height, BufferedImage.TRANSLUCENT));
		Graphics2D g2d = bufferHideBarsLayer.get().createGraphics();
		g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		g2d.drawImage(bufferLineChartLayer.get(), clip.x, clip.y, clip.x + clip.width, clip.y + clip.height,
				clip.x, clip.y, clip.x + clip.width, clip.y + clip.height, null);
		
		// add top divider
		g2d.setColor(UIColors.lineChartDividerBackground);
		g2d.fillRect(0, internalMargin - (traceBarHeight / 2), width - 1, topDivider);
		g2d.setColor(UIColors.lineChartDividerLine);
		g2d.fillRect(0, internalMargin + topDivider - dividerBarSize, width - 1, dividerBarSize);
		g2d.fillRoundRect(
				(width / 2) - dividerPointSize,
				internalMargin + topDivider - (dividerPointSize / 2),
				dividerPointSize * 2, dividerPointSize,
				10, 10);
		g2d.setColor(UIColors.lineChartDividerLine.darker());
		g2d.fillRoundRect(
				(width / 2) - dividerPointSize + 3,
				internalMargin + topDivider - 2,
				dividerPointSize * 2 - 6, 4,
				5, 5);
		
		// add bottom divider
		int dividerPosition = height - bottomDivider - internalMargin;
		
		g2d.setColor(UIColors.lineChartDividerBackground);
		g2d.fillRect(0, dividerPosition, width - 1, bottomDivider + (traceBarHeight / 2));
		g2d.setColor(UIColors.lineChartDividerLine);
		g2d.fillRect(0, dividerPosition - (dividerBarSize / 2), width - 1, dividerBarSize);
		g2d.fillRoundRect(
				(width / 2) - dividerPointSize,
				dividerPosition - (dividerPointSize / 2),
				dividerPointSize * 2, dividerPointSize,
				10, 10);
		g2d.setColor(UIColors.lineChartDividerLine.darker());
		g2d.fillRoundRect(
				(width / 2) - dividerPointSize + 3,
				dividerPosition - 2,
				dividerPointSize * 2 - 6, 4,
				5, 5);
		
		// final paint stuff
		g2d.dispose();
		
		g.drawImage(bufferHideBarsLayer.get(), clip.x, clip.y, clip.x + clip.width, clip.y + clip.height,
				clip.x, clip.y, clip.x + clip.width, clip.y + clip.height, null);
	}

	@Override
	public void mouseDragged(MouseEvent e) {
		int newY = e.getY();
		newY = (newY < internalMargin)? internalMargin : newY;
		newY = (newY > height - internalMargin)? height - internalMargin : newY;
		
		if (draggingTopBar) {
			if (newY + bottomDivider + internalMargin <= height) {
				topDivider = newY - internalMargin;
			}
		}
		
		if (draggingBottomBar) {
			if ((height - newY) + topDivider + internalMargin <= height) {
				bottomDivider = height - newY - internalMargin;
			}
		}
		
		if (draggingTopBar || draggingBottomBar) {
			logFilter.updateFrequencies(getLowerBound(), getUpperBound());
		}
		
		repaint();
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		if (mouseOverTopBar(e.getY())) {
			setCursor(Cursor.getPredefinedCursor(Cursor.S_RESIZE_CURSOR));
		} else if (mouseOverBottomBar(e.getY())) {
			setCursor(Cursor.getPredefinedCursor(Cursor.N_RESIZE_CURSOR));
		} else {
			setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		}
	}

	@Override
	public void mouseClicked(MouseEvent e) { }
	
	@Override
	public void mousePressed(MouseEvent e) {
		if (mouseOverTopBar(e.getY())) {
			draggingTopBar = true;
		} else if (mouseOverBottomBar(e.getY())) {
			draggingBottomBar = true;
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		draggingTopBar = draggingBottomBar = false;
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}

	@Override
	public void mouseEntered(MouseEvent e) { }

	@Override
	public void mouseExited(MouseEvent e) { }
	
	/**
	 * 
	 * @param y
	 * @return
	 */
	private boolean mouseOverTopBar(int y) {
		int barPosition = topDivider + internalMargin;
		return (y > barPosition - dividerPointSize) && 
				(y < barPosition + dividerPointSize);
	}
	
	/**
	 * 
	 * @param y
	 * @return
	 */
	private boolean mouseOverBottomBar(int y) {
		int barPosition = height - bottomDivider - internalMargin;
		return (y > barPosition - dividerPointSize) && 
				(y < barPosition + dividerPointSize);
	}
}
